#!C:/Programs/Strawberry/win64/perl/bin/perl.exe -I"."

###################################
#             Biology             #
###################################
# DNA molecular weight calculator #
###################################

use strict; use warnings;

use CGI;
use CGI::Carp "fatalsToBrowser";
use File::Basename;
$CGI::POST_MAX = 1024 * 25;

my $cgi = new CGI;
print "Content-Type: text/html\n\n";
print '<!DOCTYPE html>', "\n\n";
# ----------------------------------------------------------------
# Set $app_dir to the full path of the directory,
# where the template file "dna_molweight.template.html" is located
my $app_dir = "C:\\Programs\\Apache24\\htdocs";
# ----------------------------------------------------------------
my $strands = 'single'; my $topology = 'linear'; my $alldna = '';
my $filename = ''; my $fileusage = ''; my $mess = '';
my @headers = (); my @molweightsMin = (); my @molweightsMax = ();
# Get parameters from 'calling' webpage
my %params = $cgi->Vars;
my $action = lc($params{'action'});
if ($action eq 'calculate') {
  $strands = $params{'strands'}; $topology = $params{'topology'};
  $fileusage = $params{'usefile'}; $filename = $params{'filename'};
  if (($strands eq 'single' or $strands eq 'double') and ($topology eq 'linear' or $topology eq 'circular') and ($fileusage eq 'usefile' or $fileusage eq '')) {
    if ($fileusage eq 'usefile') {
      if ($filename) {
        my ( $name, $path, $extension ) = fileparse ( $filename, '..*' );
        $filename = $name . $extension; my $filename2 = $filename;
        $filename2 =~ s/[^a-zA-Z0-9 _\.\-]//g;
        if ($filename2 eq $filename) {
          my $FH = $cgi->upload("filename");
          while (<$FH>) {
            $alldna .= $_;
          }
        }
        else {
          $mess = "Invalid filename: $filename!";
        }
      }
      else {
        $mess = 'No filename or could not upload local DNA file!';
      }
    }
    else {
      $alldna = $params{'dna'};
    }
    if ($alldna) {
      my ($ref_headers, $ref_molweightsMin, $ref_molweightsMax) = calculateMolecularWeights($alldna, $strands, $topology);
      @headers = @$ref_headers; @molweightsMin = @$ref_molweightsMin; @molweightsMax = @$ref_molweightsMax;
    }
    else {
      unless ($mess) {
        $mess = 'No DNA sequence...';
      }
    }
  }
}
# Read template HTML file
my $template = "$app_dir\\dna_molweight.template.html";
open(my $input, "<", $template)
  or die "Can't open template file $template: $!";
my @lines = <$input>;
close($input);
# Create the webpage
foreach my $line (@lines) {
  chomp($line);
  if ($line) {
    if (substr($line, 0, 1) eq '#') {                                          # a '#' at beginning of line indicates that line contains a tag
      $line = substr($line, 1);                                                # remove the '#' at beginning of line
      if ($line =~ /#checked_single#/) {
        if ($strands eq 'single') {
         $line =~ s/#checked_single#/checked="checked"/;                       # check 'single stranded' radiobutton
        }
        else {
          $line =~ s/#checked_single#//;
        }
      }
      elsif ($line =~ /#checked_double#/) {
        if ($strands eq 'double') {
         $line =~ s/#checked_double#/checked="checked"/;                       # check 'double stranded' radiobutton
        }
        else {
          $line =~ s/#checked_double#//;
        }
      }
      elsif ($line =~ /#checked_linear#/) {
        if ($topology eq 'linear') {
         $line =~ s/#checked_linear#/checked="checked"/;                       # check 'linear topology' radiobutton
        }
        else {
          $line =~ s/#checked_linear#//;
        }
      }
      elsif ($line =~ /#checked_circular#/) {
        if ($topology eq 'circular') {
         $line =~ s/#checked_circular#/checked="checked"/;                     # check 'circular topology' radiobutton
        }
        else {
          $line =~ s/#checked_circular#//;
        }
      }
      elsif ($line =~ /#dna#/) {
        $line =~ s/#dna#/$alldna/;                                             # fill in the DNA sequence (or nothing)
      }
      elsif ($line =~ /#checked_usefile#/) {
        if ($fileusage eq 'usefile') {
          $line =~ s/#checked_usefile#/checked="checked"/;                     # check 'load DNA from file' checkbox
        }
        else {
          $line =~ s/#checked_usefile#//;
        }
      }
      elsif ($line =~ /#table#/) {
        # Create table HTML
        if ($action eq 'calculate') {
          my $table = '<table width="90%" border="1" style="padding:5px; margin-top: 10px">';
          # Table header
          $table .= '<tr><th style="padding:5px; align:center">&nbsp;</th><th colspan="2" style="padding:5px; text-align:center">Molecuar weight</th></tr>';
          $table .= '<tr><th style="padding:5px">DNA sequence</th><th style="padding:5px; text-align:center">Minimum</th><th style="padding:5px; text-align:center">Maximum</th></tr>';
          # Table data (one row for each sequence)
          for (my $i=0; $i < scalar(@headers); $i++) {
            $table .= '<tr><td style="padding:5px">' . $headers[$i] . '</td>';
            if ($molweightsMin[$i] == 0) {
              # If DNA sequence contains invalid base codes, the calculation sub returns $molweightsMin[$i] = 0
              $table .= '<td style="padding:5px; text-align: center" colspan="2">Invalid DNA sequence</td>';
            }
            elsif ($molweightsMin[$i] == $molweightsMax[$i]) {
              # If DNA sequence contains standard base codes only, the calculation sub returns $molweightsMin[$i] = $molweightsMax[$i]
              $table .= '<td style="padding:5px; text-align: center" colspan="2">' . sprintf("%.2f", $molweightsMin[$i]) . '</td>';
            }
            else {
              # If DNA sequence contains extended base codes, the calculation sub returns 2 different molecular weight values
              $table .= '<td style="padding:5px; text-align: right">' . sprintf("%.2f", $molweightsMin[$i]) . '</td>';
              $table .= '<td style="padding:5px; text-align: right">' . sprintf("%.2f", $molweightsMax[$i]) . '</td></tr>';
            }
          }
          $table .= '</table>';
          $line =~ s/#table#/$table/;
        }
        else {
          $line =~ s/#table#/&nbsp;/;
        }
      }
      elsif ($line =~ /#mess#/) {
         $line =~ s/#mess#/$mess/;                                             # fill in error message (or nothing, if all OK)
      }
    }
  }
  # Print file-line (with tags replaced) to webpage
  unless ($line =~ /#/) {
    print "$line\n";
  }
}

#----------------------------------------------------------------#
# Calculate molecular weight(s) for one or several DNA sequences #
#----------------------------------------------------------------#

sub calculateMolecularWeights {

  # The $alldna string may contain:
  #   - one raw data DNA sequence
  #   - one or several FASTA-formatted DNA sequences

  my ($alldna, $strands, $topology) = @_;
  my @headers = (); my @molweights1 = (); my @molweights2 = ();
  my @alldna_lines = split(/\n/, $alldna);                                     # convert string variable to array (with elements = lines, as entered in the form's textarea)
  # Check if sequence is raw data or FASTA format
  my $fasta = 0;
  if (substr($alldna_lines[0], 0, 1) eq '>') {                                 # FASTA header, starting with '>'
    $fasta = 1;
  }
  # Extract the different DNA sequences and, for each of them, calculate the molecular weight
  my $i = 0; my $j = 0; my $header = ''; my $dna = ''; my $alldone = 0;
  while (!$alldone) {
    if ($fasta) {
      # Extract FASTA header if sequences are this format
      $header = substr($alldna_lines[$i], 1);                                  # remove the '>' from FASTA header
      if (length($header) == 1) {
        # Empty FASTA header
        $header = 'Unknown DNA sequence ' . substr($alldna, 0, 12);
        if (length($alldna) > 12) {
          $header .= '...';
        }
      }
      elsif (length($header) > 35) {
        ($header) = split(/\|/, $header);                                      # use first part of FASTA header (normally DNA name)
        if (length($header) > 35) {
          $header = substr($header, 0, 32) . '...';                            # use 32 first characters of first part of FASTA header
        }
      }
      $i++;                                                                    # point to next array element (= begin of DNA sequence)
    }
    my $done = 0;
    while (!$done) {
      # If sequences are FASTA format, the '#' indicates the begin of a new sequence (thus ends the data part of actual sequence)
      if ($fasta and substr($alldna_lines[$i], 0, 1) eq '>') {
        $done = 1;
      }
      # If sequences are not FASTA format, or if the line does not start with '#', add the line content to the actual DNA sequence
      else {
        $dna .= $alldna_lines[$i];
        $i++;                                                                  # point to the next array element
      }
      # Stop parsing, if all array elements (lines of the form's textarea have been processed)
      if ($i == scalar(@alldna_lines)) {
        $done = 1; $alldone = 1;
      }
    }
    $dna =~ s/\s//g;                                                           # remove white-spaces (in particular the \n at the end of the line)
    $dna = uc($dna);
    # Save the actual FASTA header (or add some text, if is raw data)
    if ($fasta) {
      $headers[$j] = $header;
    }
    else {
      $headers[$j] = 'DNA sequence ' . substr($dna, 0, 20) . '...';
    }
    # Calculate and save the molecular weight(s) of the DNA sequence
    ($molweights1[$j], $molweights2[$j]) = molecularWeightSeq($dna, $topology);
    # If calculation has to be done for double-stranded DNA, determine the reverse-complement
    # and calculate its molecular weight(s); add this value to value from before to get total molecular weight
    if ($strands eq 'double') {
      my ($molweights_rc1, $molweights_rc2) = molecularWeightSeq(reverseComp($dna), $topology);
      $molweights1[$j] += $molweights_rc1; $molweights2[$j] += $molweights_rc2;
    }
    # Prepare for next DNA sequence
    $dna = '';
    $j++;                                                                      # point to next element in results arrays
  }
  return(\@headers, \@molweights1, \@molweights2);

}

#-----------------------------------------------#
# Calculate molecular weight(s) of DNA sequence #
#-----------------------------------------------#

sub molecularWeightSeq {

  # $DNA is a single-stranded DNA sequence
  # The sequence may contain extended base codes
  # If the sequence includes invalid characters (incl. the RNA base code U), molecular weights will be set to 0

  my ($dna, $topology) = @_; my $mw1 = 0; my $mw2 = 0;
  my $oh = 17.01;                                                              # molecular weight of 1 oxygen + 1 hydrogen
  if (validDna($dna)) {
    # If the sequence is valid DNA, calculate molecular weight(s) as sum of the bases weights
    for (my $i=0; $i < length($dna); $i++) {
      my ($bmw1, $bmw2) = molecularWeightBase(substr($dna, $i, 1));
      $mw1 += $bmw1; $mw2 += $bmw2;
    }
    # For a linear sequence, add the molecular weight of the supplementary O and H at the "free" 5' end
    if ($topology eq 'linear') {
      $mw1 += $oh; $mw2 += $oh;
    }
  }
  return($mw1, $mw2);

}

#-------------------------------------------#
# Calculate molecular weight(s) of DNA base #
#-------------------------------------------#

sub molecularWeightBase {

  # $base is a valid (standard or extended) DNA base code

  my ($base) = @_; my $mw1 = 0; my $mw2 = 0;
  # Extended to standard base code conversion; standard bases ordered by molecular weight!
  my %dna_extended = (
    'M' => 'CA',   'R' => 'AG',   'W' => 'TA',   'S' => 'CG',   'Y' => 'CT',  'K' => 'TG',
    'V' => 'CAG',  'H' => 'CTA',  'D' => 'TAG',  'B' => 'CTG',  'N' => 'CTAG'
  );
  # Molecular weights of standard DNA bases
  my %baseWeights = (
    'A' => 313.21,  'C' => 289.18,  'G' => 329.21,  'T' => 304.20
  );
  # Standard base code (A, C, G, T)
  if ($base =~ /^([ACGT])$/) {
    $mw1 = $baseWeights{$base}; $mw2 = $mw1;
  }
  # Extended base code
  else {
    my $bases_extended = $dna_extended{$base};
    $mw1 = $baseWeights{substr($bases_extended, 0, 1)}; $mw2 = $baseWeights{substr($bases_extended, -1, 1)};
  }
  return($mw1, $mw2);

}

#---------------------------------#
# DNA sequence reverse-complement #
#---------------------------------#

sub reverseComp {

  # $dna is valid DNA sequence (standard or extended base codes)

  my ($dna) = @_;
  # For each base of the sequence, determine the complementary base
  for (my $i = 0; $i < length($dna); $i++) {
    substr($dna, $i, 1, baseComp(substr($dna, $i, 1)));
  }
  # Reverse the entire sequence
  $dna = reverse($dna);
  return $dna;

}

#---------------------#
# DNA base complement #
#---------------------#

sub baseComp {

  # $base is valid DNA base (standard or extended)

  my ($base) = @_;
    $base =~ tr/ATCGMRWSYKVHDBN/TAGCKYWSRMBDHVN/;
  return $base;

}

#------------------------------#
# Check DNA sequence validity  #
#------------------------------#

sub validDna {

  my ($dna) = @_; my $valid = 0;
  my $bases = 'ACGTMRWSYKVHDBN';
  if ($dna =~ /^([$bases]+)$/) {
    $valid = 1;
  }
  return $valid;

}
